from r3frame.globals import pg
from r3frame.utils import point_inside
from r3frame.application.inputs import Mouse
from r3frame.application.resource import Window
from r3frame.application.ui.button import Button

class Interface:
    def __init__(
            self, name: str, window: Window,
            size: list[int], location: list[float],
            font_path:str, title_color:list[int]=[255, 255, 255],
            text_color:list[int]=[255, 255, 255], text_size: int=18
        ) -> None:
        pg.font.init()
        self.name = name
        self.window = window
        self.size: list[int] = size
        self.location: list[float] = location

        self.buttons = {}
        
        self.font_path = font_path
        self.text_size: int = text_size
        self.text_fields: dict[str, str] = {}
        self.text_color: list[int] = text_color
        self.title_color: list[int] = title_color
        self.font: pg.Font = pg.Font(font_path, text_size)

        self.show_name = True

    def set_button(
            self, key: str, text: str="Button",
            size: list[int]=[64, 64], location: list[int|float]=[0, 0], title_color:list[int]=[255, 255, 255],
            color: list[int]=[0, 0, 0], text_color:list[int]=[255, 255, 255], text_size: int=18, padding: list[int]=[0, 0],
            border_size: list[int]=[5, 5], border_radius: list[int]=[0, 0, 0, 0], border_color: list[int]=[255, 255, 255],
        ) -> None:
        self.buttons[key] = Button(
            self.font_path, text, size, location,
            color, text_color, text_size, padding,
            border_size, border_radius, border_color
        )
    
    def get_button(self, key: str) -> Button|None:
        return self.buttons.get(key, None)
    
    def rem_button(self, key: str) -> Button|None:
        if self.get_button(key) is not None:
            del self.buttons[key]

    def set_text_field(self, field: str, text: str, color: list[int]=None) -> bool:
        try:
            self.text_fields[field] = {"text": text, "color": color}
            return True
        except KeyError as e:
            return False
    
    def rem_text_field(self, field: str) -> bool:
        try:
            del self.text_fields[field]
            return True
        except KeyError as e:
            return False

    def render(self) -> None:
        for button in self.buttons.values():
            button.render(self.window.window)
        
        if self.show_name: self.window.window.blit(self.font.render(self.name, True, self.title_color), self.location)
        
        for index, field in enumerate(self.text_fields.keys()):
            text = f"{field}: {self.text_fields[field]["text"]}"
            color = self.text_fields[field]["color"]
            text_surface = self.font.render(text, True, self.text_color if not color else color)
            text_location = [
                self.location[0],
                self.location[1] + (text_surface.get_size()[1] * (index + 1))
            ]
            self.window.window.blit(text_surface, text_location)

    def update(self, event_manager) -> None:
        for button in self.buttons.values():
            mouse_location = Mouse.get_location()
            if not button.hovered and point_inside(mouse_location, [*button.location, *button.size]):
                Mouse.Hovering = button
                button.hovered = True
                button.on_hover()
            if button.hovered and not point_inside(mouse_location, [*button.location, *button.size]):
                Mouse.Hovering = None
                button.hovered = False
                button.on_unhover()
            if button.hovered and event_manager.mouse_pressed(Mouse.LeftClick):
                button.on_click()



from r3frame.application.ui import Button
from r3frame.application.scene import Scene
from r3frame.application.events import Event_Manager
from r3frame.application.inputs import Keyboard, Mouse, Action_Map
from r3frame.application.resource import Clock, Window, Camera, Renderer

class Application:
    def __init__(self, name: str="My App") -> None:
        self.name = name
        self.clock = Clock()
        self.events = Event_Manager()

        self.scene: str = None
        self.active_scene: Scene = None
        self.scenes: dict[str, Scene] = {}

    def set_scene(self, scene: Scene) -> None:
        self.scenes[scene.name] = scene
        self.scene = scene.name
        self.active_scene = self.scenes[self.scene]
    def rem_scene(self, key: str) -> Scene|None:
        if self.get_scene(key) is not None:
            del self.scenes[key]
            self.scene = None
            self.active_scene = None
    def get_scene(self, key: str) -> Scene|None: return self.scenes.get(key, None)

    def load_assets(self) -> None: raise NotImplementedError
    def load_scenes(self) -> None: raise NotImplementedError
    def load_objects(self) -> None: raise NotImplementedError

    def handle_events(self) -> None: raise NotImplementedError
    def handle_update(self) -> None: raise NotImplementedError
    def handle_render(self) -> None: raise NotImplementedError

    def run(self) -> None:
        self.load_scenes()
        self.load_assets()
        self.load_objects()
        while not self.events.quit:
            self.clock.update()
            self.events.update()

            self.handle_events()

            self.handle_update()
            if self.scene is not None:
                self.scenes[self.scene].camera.update(self.clock.delta)
                self.scenes[self.scene].handle_update(self.events)

            self.handle_render()
            if self.scene is not None:
                self.scenes[self.scene].renderer.flush()
                self.scenes[self.scene].handle_render()
                self.scenes[self.scene].window.update()

            self.clock.rest()


import r3frame.application.ui as ui
import r3frame.application.scene as scene
import r3frame.application.inputs as inputs
import r3frame.application.events as events
import r3frame.application.resource as resource

from r3frame.application.base import Application



import r3frame as r3
import random

class Playground(r3.app.Application):
    def __init__(self):
        super().__init__("r3 playground")

    def load_assets(self) -> None:
        self.active_scene.assets.load_image("link", r3.utils._asset_path("images/link.png"), [16, 16])
        self.active_scene.assets.load_image("green-link", r3.utils._asset_path("images/green-link.png"), [16, 16])
        self.active_scene.assets.load_image_sheet("run", r3.utils._asset_path("images/run.png"), [32, 32])
        self.active_scene.assets.load_image_sheet("idle", r3.utils._asset_path("images/idle.png"), [32, 32])
        
    def load_scenes(self) -> None:
        self.set_scene(r3.app.scene.Scene("r3 playground", [800, 600], r3.objects.world.Grid_Map(150, 150, 16)))
        
        self.active_scene.set_interface(r3.app.ui.Interface("Debug", self.active_scene.window, [100, 100], [0, 0], r3.utils._asset_path("fonts/megamax.ttf")))
        self.active_scene.set_interface(r3.app.ui.Interface("Buttons", self.active_scene.window, [200, 50], [5, 550], r3.utils._asset_path("fonts/megamax.ttf")))
        self.active_scene.set_interface(r3.app.ui.Interface("Controls", self.active_scene.window, [200, 50], [600, 5], r3.utils._asset_path("fonts/megamax.ttf")))
        
        for i in range(2):
            text = ["Hello!", "Debug"]
            self.active_scene.get_interface("Buttons").set_button(f"button-{i+1}", text[i], [200, 50], location=[205 * i, 550], padding=[65, 16], text_size=16)

        self.active_scene.get_interface("Buttons").show_name = False
        button_1 = self.active_scene.get_interface("Buttons").get_button("button-1")
        def on_click() -> None:
            button_1.text = random.choice(["Just Wow!", "Help Me!", "Hahaha!", "Not Funny!", "The Colors!", "Ouch!", "The Pain!", "r3frame!", "I Get It!", "OK!"])
            button_1.border_color = button_1.text_color = [
                random.randint(1, 255),
                random.randint(1, 255),
                random.randint(1, 255)
            ]
            button_1.padding = [45, 16]
        def on_hover() -> None:
            button_1.text = "Click Me!"
            button_1.padding = [45, 16]
            button_1.border_color = button_1.text_color = [0, 255, 0]
        def on_unhover() -> None:
            button_1.text = "Hello!"
            button_1.padding = [65, 16]
            button_1.border_color = button_1.text_color = [255, 255, 255]
        button_1.on_click = on_click
        button_1.on_hover = on_hover
        button_1.on_unhover = on_unhover
        
        self.debug_mode = False
        button_2 = self.active_scene.get_interface("Buttons").get_button("button-2")
        button_2.text_color = [255, 0, 0] if not self.debug_mode else [0, 255, 0]
        button_2.border_color = [255, 0, 0] if not self.debug_mode else [0, 255, 0]
        def on_click2() -> None:
            self.debug_mode = not self.debug_mode
            button_2.text_color = [255, 0, 0] if not self.debug_mode else [0, 255, 0]
            button_2.border_color = [255, 0, 0] if not self.debug_mode else [0, 255, 0]
        def on_hover2() -> None:
            button_2.border_color = [255, 255, 255]
        def on_unhover2() -> None:
            button_2.border_color = [255, 0, 0] if not self.debug_mode else [0, 255, 0]
        button_2.on_click = on_click2
        button_2.on_hover = on_hover2
        button_2.on_unhover = on_unhover2
    
    def load_objects(self) -> None:
        self.player = r3.objects.game.Game_Object(
            mass=500, speed=150,
            location=[100, 100], color=[0, 0, 255],
        )
        self.player.set_animation("idle", self.active_scene.assets.get_image("idle"), frame_duration=4)
        self.player.set_animation("run", self.active_scene.assets.get_image("run"), frame_duration=5)

        self.chain = r3.objects.game.Chain(r3.objects.game.Game_Object(color=[0, 255, 0], size=[16, 16], location=[100, 100]), [r3.objects.game.Game_Object(size=[16, 16], color=[255, 0, 0] if _+1 == 10 else [255, 255, 255]) for _ in range(10)], link_offset=[0, -8])
        for link in self.chain.links:
            link.set_image(self.active_scene.assets.get_image("link"))
        self.chain.root.set_image(self.active_scene.assets.get_image("green-link"))
        self.chain.get_scaling = lambda index, parent, link: 20
        self.chain.grabbed = False

        def post_render():
            if self.debug_mode:
                self.active_scene.get_interface("Debug").name = "Debug - Active"
                self.chain.debug_render(self.active_scene.renderer)
                self.active_scene.partition.debug_render(self.active_scene.renderer, [*map(int, [self.player.location[0] + self.player.size[0] / 2, self.player.location[1] + self.player.size[1] / 2])])
                self.active_scene.get_interface("Debug").title_color = [0, 255, 0]
                self.active_scene.renderer.set_flag(self.active_scene.renderer.FLAGS.SHOW_CAMERA)
                self.active_scene.get_interface("Debug").set_text_field("FPS", f"{self.clock.FPS:0.1f}", color=[0, 255, 0] if self.clock.FPS > self.clock.maxFPS/2 else [255, 0, 0])
                self.active_scene.get_interface("Debug").set_text_field("OBJECTS", f"{len([*self.active_scene.partition.cells, self.player, *self.chain.get_objects()])}")
                self.active_scene.get_interface("Debug").set_text_field("DIRECTION", f"{self.player.get_facing()}")
                self.active_scene.get_interface("Debug").set_text_field("VELOCITY", f"{self.player.velocity[0]:0.1f}, {self.player.velocity[1]:0.1f}")
                self.active_scene.get_interface("Debug").set_text_field("LOCATION", f"{self.player.location[0]:0.1f}, {self.player.location[1]:0.1f}")
            else:
                self.active_scene.get_interface("Debug").name = "Debug - Inactive"
                self.active_scene.get_interface("Debug").title_color = [255, 0, 0]
                self.active_scene.renderer.rem_flag(self.active_scene.renderer.FLAGS.SHOW_CAMERA)
                self.active_scene.get_interface("Debug").rem_text_field("FPS")
                self.active_scene.get_interface("Debug").rem_text_field("OBJECTS")
                self.active_scene.get_interface("Debug").rem_text_field("DIRECTION")
                self.active_scene.get_interface("Debug").rem_text_field("VELOCITY")
                self.active_scene.get_interface("Debug").rem_text_field("LOCATION")
        self.active_scene.renderer.post_render = post_render

    def handle_events(self) -> None:
        zoom_factor = 3

        if self.events.key_pressed(r3.app.inputs.Keyboard.Escape): self.events.quit = 1

        if self.events.mouse_wheel_up:
            self.active_scene.camera.mod_viewport(-zoom_factor)

        if self.events.mouse_wheel_down:
            self.active_scene.camera.mod_viewport(zoom_factor)

        if self.events.key_held(r3.app.inputs.Keyboard.E):
            mouse_location = r3.app.inputs.Mouse.get_location()
            wx, wy = [
                mouse_location[0] / self.active_scene.camera.viewport_scale[0] + self.active_scene.camera.location[0],
                mouse_location[1] / self.active_scene.camera.viewport_scale[1] + self.active_scene.camera.location[1],
            ]
            if r3.utils.point_inside([wx, wy], [*self.chain.root.location, *self.chain.root.size]):
                self.chain.grabbed = True
        else: self.chain.grabbed = False

        if not isinstance(r3.app.inputs.Mouse.Hovering, r3.app.ui.Button) and self.events.mouse_held(r3.app.inputs.Mouse.LeftClick):
            mouse_location = r3.app.inputs.Mouse.get_location()
            world_x = (mouse_location[0] / self.active_scene.camera.viewport_scale[0] + self.active_scene.camera.location[0])
            world_y = (mouse_location[1] / self.active_scene.camera.viewport_scale[1] + self.active_scene.camera.location[1])
            self.active_scene.partition.set_cell(world_x, world_y, r3.objects.game.Game_Object(
                size=[self.active_scene.partition.cell_size, self.active_scene.partition.cell_size],
                color=[random.randint(1, 255), random.randint(1, 255), random.randint(1, 255)]
            ))

        if not isinstance(r3.app.inputs.Mouse.Hovering, r3.app.ui.Button) and self.events.mouse_held(r3.app.inputs.Mouse.RightClick):
            mouse_location = r3.app.inputs.Mouse.get_location()
            world_x = (mouse_location[0] / self.active_scene.camera.viewport_scale[0] + self.active_scene.camera.location[0])
            world_y = (mouse_location[1] / self.active_scene.camera.viewport_scale[1] + self.active_scene.camera.location[1])
            self.active_scene.partition.rem_cell(world_x, world_y)

        if self.events.key_held(r3.app.inputs.Keyboard.W):       self.player.set_velocity(vy=-self.player.speed)
        if self.events.key_held(r3.app.inputs.Keyboard.S):       self.player.set_velocity(vy=self.player.speed)
        if self.events.key_held(r3.app.inputs.Keyboard.D):
            self.player.animation.flip_x =False
            self.player.set_velocity(vx=self.player.speed)
        if self.events.key_held(r3.app.inputs.Keyboard.A):
            self.player.animation.flip_x = True
            self.player.set_velocity(vx=-self.player.speed)

    def handle_update(self) -> None:
        self.chain.update(self.clock.delta)
        if self.chain.grabbed:
            mouse_location = r3.app.inputs.Mouse.get_location()
            self.chain.root.location = [
                mouse_location[0] / self.active_scene.camera.viewport_scale[0] + self.active_scene.camera.location[0],
                mouse_location[1] / self.active_scene.camera.viewport_scale[1] + self.active_scene.camera.location[1],
            ]
        if self.player.velocity[0] != 0 or self.player.velocity[1] != 0:
            self.player.set_action("run")
        else:
            self.player.set_action("idle")

        self.player.update(self.clock.delta)
        for obj in self.active_scene.partition.cells.values():
            obj.update(self.clock.delta)

        self.active_scene.camera.center_on(self.player.size, self.player.location)

    def handle_render(self) -> None:
        [self.active_scene.renderer.draw_call(o.image, o.location) for o in [*self.active_scene.partition.cells.values(), self.player, *self.chain.get_objects()]]
        self.active_scene.get_interface("Controls").set_text_field("Up", "W", [200, 200, 200])
        self.active_scene.get_interface("Controls").set_text_field("Down", "S", [200, 200, 200])
        self.active_scene.get_interface("Controls").set_text_field("Left", "A", [200, 200, 200])
        self.active_scene.get_interface("Controls").set_text_field("Right", "D", [200, 200, 200])
        self.active_scene.get_interface("Controls").set_text_field("Interact", "E", [200, 200, 200])
        self.active_scene.get_interface("Controls").set_text_field("Place", "L-Click", [200, 200, 200])
        self.active_scene.get_interface("Controls").set_text_field("Remove", "R-Click", [200, 200, 200])
        
def playground(): Playground().run()



from r3frame.objects.game.base import Game_Object

class Chain:
    def __init__(self, root: Game_Object, links: list[Game_Object], link_offset: list[int | float] = [0, 0]) -> None:
        self.root = root
        self.links = links
        self.link_offset = link_offset
        self.link()

    def link(self) -> None:
        """Set link positions based on root."""
        prev = self.root
        for link in self.links:
            link.location = [prev.location[0] + self.link_offset[0], prev.location[1] + self.link_offset[1]]
            prev = link

    def set_link(self, obj: Game_Object) -> None:
        """Sets an object to the end of the chain."""
        self.links.append(obj)
        prev = self.links[-1] if self.links else self.root
        obj.location = [prev.location[0] + self.link_offset[0], prev.location[1] + self.link_offset[1]]

    def rem_link(self) -> Game_Object|None:
        """Remove and return an object from the end of the chain."""
        if self.links:
            obj = self.links.pop()
            self.links.remove(obj)
            return obj
        else: return None

    def get_objects(self) -> list[Game_Object]:
        """ Returns a list of all objects in the chain, including the root."""
        return [self.root, *self.links]

    @staticmethod
    def get_scaling(index: int, parent: Game_Object, link: Game_Object) -> float:
        """Default scaling factor applied to link velocity updates. Override this for different chain behaviors."""
        return 10 / (index + 1)

    def update(self, delta_time: float) -> None:
        """ Updates each member of the chain using inverse kinematics."""
        self.root.update(delta_time)
        for i, link in enumerate(self.links):
            parent = self.root if i == 0 else self.links[i - 1]
            link.velocity = [
                (parent.location[0] - (link.location[0] + self.link_offset[0])) * self.get_scaling(i, parent, link),
                (parent.location[1] - (link.location[1] + self.link_offset[1])) * self.get_scaling(i, parent, link),
            ]
            link.update(delta_time)

    def debug_render(self, renderer) -> None:
        """ Renders a debug visualization of the chain. """
        prev = self.root
        for link in self.links:
            renderer.draw_circle([link.location[0] + link.size[0] / 2 , link.location[1] + link.size[1] / 2], link.size[0], [255, 255, 255], 1)
            renderer.draw_line(
                [prev.location[0] + prev.size[0] / 2, prev.location[1] + prev.size[1] / 2],
                [link.location[0] + link.size[0] / 2, link.location[1] + link.size[1] / 2], 
                [255, 0, 0], 3
            )
            prev = link
